/**********************************************************  Scanbus  *********************************************************

	Red Rapids
	797 North Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The ScanBus utility reports bus information for all Red Rapids devices detected.
	Note:  This is a simple application example for the Adapter driver and API.

	Rev History:  09-16-2015:	Added Adapter_Driver() function call to replace default driver name.
				  06-20-2011:	Inital release.

	Copyright 2015, Red Rapids, Inc.
	
**********************************************************  Scanbus  *********************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "adapter.h"

int main(int argc, char* argv[])
{
	s_Adapter *p_HW[MAX_NUM_DEVICES];	// Device structure attached to the hardware handle.
	unsigned long devcnt;				// Loop counter for device number.
	unsigned long unitcnt = 0;			// Total number of units expected.
	unsigned long opencnt = 0;			// Total number of devices successfully opened.
	unsigned long status = 0;			// Function return status.

	// Set the driver name before attempting any communication with the hardware.
	Adapter_Driver("RRadapter");

	// Initialize the hardware handle for each device.
	for (devcnt = 0; devcnt < MAX_NUM_DEVICES; devcnt++)
	{
		// Allocate memory for the structure.
		p_HW[devcnt] = (s_Adapter*) malloc(sizeof(s_Adapter));

		// Initialize each member of the structure.
		Adapter_Zero(p_HW[devcnt]);

		// Set a unique device number for each hardware handle.
		p_HW[devcnt]->DevNum = devcnt;
	}

//**************************************  Device Search  ***************************************
//  The following code searches for the maximum number of Red Rapids devices on the bus.
//**********************************************************************************************

	// Loop through the maximum number of possible devices.
	for(devcnt = 0; devcnt < MAX_NUM_DEVICES; devcnt++)
	{

		// Attempt to open each device.
		printf("\nAttempting to open device number %lu ... ", p_HW[devcnt]->DevNum);
		status = Adapter_Open(p_HW[devcnt]);
		
		// Device open succeeded.
		if (status == 0)
		{
			// Report bus information if the device opens successfully.
			printf("device found.\nBus information for device number %lu:\n", p_HW[devcnt]->DevNum);
			printf("\tBus number: %lu \n", p_HW[devcnt]->Bus);
			printf("\tSlot number: %lu \n", p_HW[devcnt]->Slot);
			printf("\tFunction number: %lu \n", p_HW[devcnt]->Function);
			// Store the reported number of units found when the first device was opened.
			if (devcnt == 0)
				unitcnt = p_HW[devcnt]->UnitsFound;
			// Increment the counter for successfully opened device.
			opencnt++;
			// Close the device.
			Adapter_Close(p_HW[devcnt]);
		}
		// Device open failed.
		else
			// Report that the device did not open.
			printf("device not found.\n");

	}

	printf("\nSuccessfully opened %lu device(s) of the %lu device(s) expected.\n", opencnt, unitcnt);

	// Free the memory allocated to the device structures.
	for (devcnt = 0; devcnt < MAX_NUM_DEVICES; devcnt++)
		free(p_HW[devcnt]);

	return 0;
}
