/**********************************************************************************************
	Red Rapids
	797 N. Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	Description: Adapter functions that are common to all products.

***********************************************************************************************/

#ifndef _ADAPTER_H_
#define _ADAPTER_H_

#if defined(__cplusplus)
    extern "C" {
#endif

// Define error codes.
#define OK					0x0
#define DRIVER_FAIL			0x10
#define OPEN_FAIL			0x20
	#define KERNEL_HANDLE		(OPEN_FAIL|0x1)	
	#define PCI_SCAN			(OPEN_FAIL|0x2)	
	#define NO_DEV_FOUND		(OPEN_FAIL|0x3)
	#define DEV_OUT_OF_RANGE	(OPEN_FAIL|0x4)
	#define PCI_INFO			(OPEN_FAIL|0x5)
	#define HANDLE_ATTACH		(OPEN_FAIL|0x6)
	#define MEM_ALLOCATE		(OPEN_FAIL|0x7)
	#define INT_ENABLE			(OPEN_FAIL|0x8)
#define CLOSE_FAIL			0x30
#define RDWR_FAIL			0x40
	#define RD_FAIL				(RDWR_FAIL|0x1)	
	#define WR_FAIL				(RDWR_FAIL|0x2)	
#define DMA_FAIL			0x50
	#define DMA_ALLOCATE		(DMA_FAIL|0x1)	
	#define DMA_FREE			(DMA_FAIL|0x2)
#define SLEEP_FAIL			0x60
 
// Define operating ranges.
#define MAX_NUM_DEVICES			8
#define MAX_NUM_DMA_CHANNELS	8
#define MAX_DMA_PAGES_PER_CH	1024

typedef struct {
	unsigned long ChannelIndex;							// 0-7
	unsigned long BurstSizeB[MAX_NUM_DMA_CHANNELS];		// Burst size in bytes.
	unsigned long BurstCount[MAX_NUM_DMA_CHANNELS];		// Number of bursts per page.
	unsigned long PageCount[MAX_NUM_DMA_CHANNELS];		// Number of pages.
	unsigned long PagesPerMark[MAX_NUM_DMA_CHANNELS];	// This variable represents the number of pages to transfer between markers (status bit, interrupt, etc.).
														// The mark is not generated by the driver, this variable is provided for convenience to the application.
	void	 *VirtDMAAdr[MAX_NUM_DMA_CHANNELS][MAX_DMA_PAGES_PER_CH];		// Virtual address of each DMA buffer for software access.
	unsigned long long PhysDMAAdr[MAX_NUM_DMA_CHANNELS][MAX_DMA_PAGES_PER_CH];		// Physical address of each DMA buffer for hardware access.
} s_DMAConfig;

typedef struct {
	unsigned long BAR;				// Global interrupt mask BAR.		
	unsigned long Offset;			// Global interrupt mask offset.
	unsigned long long Disable;		// Global interrupt mask value to disable interrupts.
} s_INTConfig;

typedef struct {
	unsigned long DevNum;			// Device Number on the PCI bus (zero based).
	unsigned long Bus;				// PCI bus number reported by the driver (zero based).
	unsigned long Slot;				// PCI slot number reported by the driver (zero based).
	unsigned long Function;			// PCI function number reported by the driver (zero based).
	unsigned long UnitsFound;		// Number of Adapters found during open() function.
	s_DMAConfig DMA;				// DMA config for all DMA buffers.
	s_INTConfig GlobalIntMask;		// Global interrupt mask structure.
	void *User;						// Undefined pointer available to user applications.
	int Verbose;					// Reserved for status message reporting.
} s_Adapter;

#ifdef WIN32
	#define EXPORTULONG __declspec(dllexport) unsigned long
#else
	#define EXPORTULONG unsigned long
#endif	

EXPORTULONG Adapter_Zero(s_Adapter *p_Adapter);
EXPORTULONG Adapter_Driver(const char* Name);

EXPORTULONG Adapter_Open(s_Adapter *p_Adapter);
EXPORTULONG Adapter_Close(s_Adapter *p_Adapter);

EXPORTULONG Adapter_Read32(s_Adapter *p_Adapter, unsigned long BAR, unsigned long Offset, unsigned *pData);
EXPORTULONG Adapter_Read64(s_Adapter *p_Adapter, unsigned long BAR, unsigned long Offset, unsigned long long *pData);

EXPORTULONG Adapter_Write32(s_Adapter *p_Adapter, unsigned long BAR, unsigned long Offset, unsigned Data);
EXPORTULONG Adapter_Write64(s_Adapter *p_Adapter, unsigned long BAR, unsigned long Offset, unsigned long long Data);

EXPORTULONG Adapter_DMABufAllocate(s_Adapter *p_Adapter);
EXPORTULONG Adapter_DMABufFree(s_Adapter *p_Adapter);

EXPORTULONG Adapter_uSleep(unsigned long uSec);

#ifdef __cplusplus
}
#endif

#endif

